__all__ = [
    "NodeMessageSession",
    "HostMessageSession",
    "MessageSession",
    "SessionIsBroken",
    "Message",
    "Event",
    "Payload",
    "PayloadEncoder",
    "EncodingError",
    "MQTT5Client",
    "MQTTMessage",
    "MQTTError",
]


from .hosts import HostMessageSession
from .nodes import NodeMessageSession
from .ports import (
    EncodingError,
    MessageSession,
    MQTT5Client,
    MQTTError,
    PayloadEncoder,
    SessionIsBroken,
)
from .values import Event, Message, MQTTMessage, Payload



__all__ = (
    "Host",
    "CommandRequest",
    "CommandFollower",
    "MetricsCreator",
    "MinMetricInfoCache",
    "MinMetricInfo",
    "WriteRequest",
)

from .commands import MinMetricInfo, MinMetricInfoCache
from .hosts import CommandFollower, CommandRequest, Host, MetricsCreator
from .values import WriteRequest



import asyncio
import logging
from typing import Callable, Coroutine

from aiospb.messages import Event, Message, MessageSession
from aiospb.messages.ports import SessionIsBroken
from aiospb.shared import Clock, UtcClock

from ..shared import DataType, DeviceKey, HostOptions, Metric
from .commands import CommandFollower, MetricsCreator
from .values import CommandRequest, CommandResponse, WritingResolution

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


_CTR_NAMES = (
    "Node Control/Rebirth",
    "Node Control/Restart",
    "Node Control/Reboot",
    "Device Control/Rebirth",
    "Device Control/Restart",
    "Device Control/Reboot",
)


class Host:
    """Host to notify messages to observers and execute commands"""

    def __init__(
        self,
        key: DeviceKey,
        host_session: MessageSession,
        metrics_creator: MetricsCreator,
        command_follower: CommandFollower,
        options: HostOptions,
        clock: Clock | None = None,
    ):
        self._key = key
        self._session = host_session
        self._creator = metrics_creator
        self._follower = command_follower
        self._opts = options

        self._clock = clock or UtcClock()

        self._observers = {}
        self._incomming = None

    async def _subscribe_to_nodes(self):
        try:
            async for msg in self._session.subscribe():
                if msg.event == Event.OUT_OF_SEQUENCE:
                    logger.warning(
                        f"Found out of sequence at {msg.key}, sending rebirth..."
                    )
                    await self._session.publish(
                        Message(
                            Event.NODE_COMMAND_IS_SENT,
                            msg.key,
                            [
                                Metric(
                                    msg.sending_ts,
                                    True,
                                    DataType.Boolean,
                                    name="Node Control/Rebirth",
                                )
                            ],
                        )
                    )
                    continue

                await self._notify_to_observers(msg)

        except SessionIsBroken:
            await self._notify_to_observers(
                Message(Event.HOST_STATE_HAS_CHANGED, self._key, online=False)
            )
            self._incomming = None
        except asyncio.CancelledError:
            if self._incomming:
                self._incomming.cancel()
                try:
                    await self._incomming
                except asyncio.CancelledError:
                    ...
        except Exception as e:
            logger.warning("Error while procesing message")
            logger.exception(e)

            if self._incomming:
                self._incomming.cancel()
                try:
                    await self._incomming
                except asyncio.CancelledError:
                    ...
            self._incomming = asyncio.create_task(self._subscribe_to_nodes())
            raise e

    async def _notify_to_observers(self, msg: Message):
        timeout = self._opts.notify_timeout
        tasks = [
            asyncio.create_task(callback(msg), name=key)
            for key, callback in self._observers.items()
        ]
        _, pending = await asyncio.wait(tasks, timeout=timeout)
        for t in pending:
            logger.warning(
                f'Notification of {msg.event} from "{msg.key.path}" to {t.get_name()} is cancelled'
            )
            t.cancel()
            # try:
            #     await t
            # except asyncio.CancelledError:
            #     ...

    def with_options(self, value: HostOptions):
        self._opts = value

    async def establish_session(self):
        if self._incomming:
            self._incomming.cancel()
            self._incomming = None

        await self._session.establish()

        if "follower" not in self._observers:
            self.add_observer("follower", self._follower.update_command_pipes)

        if "creator" not in self._observers:
            self.add_observer("creator", self._creator.update_metric_pars)

        self._incomming = asyncio.create_task(self._subscribe_to_nodes())

    async def terminate_session(self):
        await self._terminate()
        await self._session.terminate()

    async def execute_command(self, req: CommandRequest) -> CommandResponse:
        if not self._session.is_established():
            return CommandResponse(
                self._clock.timestamp(),
                [WritingResolution.HostIsOffline] * len(req.requests),
            )

        metrics = await self._creator.construct_writing_metrics(req)
        if DataType.Unknown in [m.data_type for m in metrics]:
            return CommandResponse(
                self._clock.timestamp(),
                [
                    WritingResolution.MetricNotFound
                    if m.data_type == DataType.Unknown
                    else WritingResolution.Cancelled
                    for m in metrics
                ],
            )
        event = (
            Event.DEVICE_COMMAND_IS_SENT
            if req.key.device_name
            else Event.NODE_COMMAND_IS_SENT
        )

        names = [m.name for m in metrics]
        for cname in _CTR_NAMES:
            if cname in names:
                if len(names) > 1:
                    logger.error(
                        f"Not possible to send control command {cname} with other metrics"
                    )
                    return CommandResponse(
                        self._clock.timestamp(),
                        [WritingResolution.BadWriting] * len(names),
                    )

                try:
                    await self._session.publish(Message(event, req.key, metrics))
                    return CommandResponse(
                        self._clock.timestamp(), [WritingResolution.GoodWriting]
                    )
                except Exception as e:
                    logger.error(f"Failed to send control {cname} to {req.key}")
                    logger.exception(e)
                    return CommandResponse(
                        self._clock.timestamp(), [WritingResolution.BadWriting]
                    )

        try:
            await self._session.publish(Message(event, req.key, metrics))
        except SessionIsBroken:
            return CommandResponse(
                self._clock.timestamp(),
                [WritingResolution.HostIsOffline] * len(metrics),
            )

        return await self._follower.follow_command(
            req.key, metrics, req.writing_timeout
        )

    def add_observer(
        self,
        key: str,
        callback: Callable[[Message], Coroutine[None, None, None]],
    ):
        self._observers[key] = callback

    async def _terminate(self):
        if self._incomming:
            self._incomming.cancel()
            try:
                await self._incomming
            except asyncio.CancelledError:
                ...
            self._incomming = None



import abc

from ..shared import DeviceKey, Metric, ValueType
from .values import MetricInfo, Reading


class Historian(abc.ABC):
    @abc.abstractmethod
    async def save(self, key: DeviceKey, metrics: list[Metric]):
        """Persist the metrics"""

    @abc.abstractmethod
    async def load(self, key: DeviceKey) -> list[Metric]:
        """Return the persisted metrics, but marked as is_history"""

    @abc.abstractmethod
    async def clear(self, key: DeviceKey):
        """Clear all the history of a device (beacuse it was succesfully sent)"""

    @abc.abstractmethod
    def get_device_keys(self, node_key: DeviceKey) -> list[DeviceKey]:
        """List all device_keys with stored histories related to node_key"""


class ReadingStore(abc.ABC):
    @abc.abstractmethod
    async def get(self, ids: list[int | str]) -> list[Reading]:
        """Return the values of the aliases listed"""

    @abc.abstractmethod
    async def set(self, readings: list[Reading]):
        """Set a group of readings to some aliases"""


class MetricNotFound(Exception):
    def __init__(self, name: str, alias: int):
        self.name = name
        self.alias = alias
        super().__init__(f'Metric "{name}:{alias}" not found in MetricsNetwork')


class DeviceConnectionIsBroken(Exception):
    def __init__(self, key: DeviceKey, conn_name: str = ""):
        self.device_key = key
        self.conn_name = conn_name
        super().__init__(f"Connection {conn_name} to device {key} has broken")


class MetricsNetwork(abc.ABC):
    @abc.abstractmethod
    def get_device_keys(self) -> list[DeviceKey]:
        ...

    @abc.abstractmethod
    async def get_info(self, key: DeviceKey) -> list[MetricInfo]:
        """List the metric information available for a device"""
        ...

    @abc.abstractmethod
    async def read(self, key: DeviceKey, alias: int = 0, name: str = "") -> ValueType:
        """Read the current value of a metric, defined by alias or name"""

    @abc.abstractmethod
    async def write(
        self, key: DeviceKey, value: ValueType, alias: int = 0, name: str = ""
    ) -> int:
        """Write the value on the metric defined by the alias, returns it alias as confirmation"""

    @abc.abstractmethod
    async def connect_device(self, key: DeviceKey):
        """Connect device"""

    @abc.abstractmethod
    async def disconnect_device(self, key: DeviceKey):
        """Disconnect device"""



import abc
from typing import AsyncIterable, AsyncIterator

from .values import DeviceKey, Event, Message, MQTTMessage, Payload


class SessionIsBroken(Exception):
    """MessageSession has broken accidentallly"""


class MessageSession(abc.ABC):
    """Interface to manage messages, plain msgs are the only value objects it manages"""

    @abc.abstractmethod
    async def establish(self):
        ...

    @abc.abstractmethod
    async def terminate(self):
        ...

    @abc.abstractmethod
    def is_established(self) -> bool:
        ...

    @abc.abstractmethod
    async def subscribe(self) -> AsyncIterable[Message]:
        yield Message(Event.NODE_HAS_DIED, DeviceKey(""))

    @abc.abstractmethod
    async def publish(self, msg: Message):
        ...


class MQTTError(Exception):
    """Connection error to MQTT Broker, it implies clien is not connected"""

    def __init__(self, wrapped: Exception):
        self.wrapped = wrapped
        super().__init__(str(wrapped))


class MQTT5Client(abc.ABC):
    """Interface to publish messages to MQTT"""

    @abc.abstractmethod
    async def connect(self, will: MQTTMessage):
        """Connect to MQTTBroker"""

    @abc.abstractmethod
    async def publish(self, message: MQTTMessage):
        """Publish message to MQTT broker, it disconnect from broker if MQTTError"""

    @abc.abstractmethod
    async def disconnect(self):
        """Disconnect from the MQTTBroker, it shouldn't raise any exception..."""

    @abc.abstractmethod
    async def subscribe(self, wildcard: str, qos: int):
        """Subscribe to recieve MQTT broker messages, it disconnect from broker if MQTTError"""

    @abc.abstractmethod
    async def unsubscribe(self, wildcard: str):
        """Unsubscribe to recieve MQTT broker messages, it disconnect from broker if MQTTError"""

    @property
    @abc.abstractmethod
    def messages(self) -> AsyncIterator[MQTTMessage]:
        """Messages subscribed are recevieved in a async for"""

    @abc.abstractmethod
    def is_connected(self) -> bool:
        """Check if the client is connected to MQTT broker"""


class EncodingError(Exception):
    """Failure when encoding/decoding the payload of MQTTMessage"""


class PayloadEncoder(abc.ABC):
    """Encode the message before sending it as payload in the message"""

    @abc.abstractmethod
    def encode(self, payload: Payload) -> bytes:
        """Convert a payload to a sequence of bytes"""
        ...

    @abc.abstractmethod
    def decode(self, b: bytes) -> Payload:
        """Convert bytes to a payload class"""



from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import python_message as _python_message

DESCRIPTOR: _descriptor.FileDescriptor

class Payload(_message.Message):
    __slots__ = ("timestamp", "metrics", "seq", "uuid", "body")
    Extensions: _python_message._ExtensionDict

    class Template(_message.Message):
        __slots__ = (
            "version",
            "metrics",
            "parameters",
            "template_ref",
            "is_definition",
        )
        Extensions: _python_message._ExtensionDict

        class Parameter(_message.Message):
            __slots__ = (
                "name",
                "type",
                "int_value",
                "long_value",
                "float_value",
                "double_value",
                "boolean_value",
                "string_value",
                "extension_value",
            )

            class ParameterValueExtension(_message.Message):
                __slots__ = ()
                Extensions: _python_message._ExtensionDict
                def __init__(self) -> None: ...
            NAME_FIELD_NUMBER: _ClassVar[int]
            TYPE_FIELD_NUMBER: _ClassVar[int]
            INT_VALUE_FIELD_NUMBER: _ClassVar[int]
            LONG_VALUE_FIELD_NUMBER: _ClassVar[int]
            FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
            DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
            BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
            STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
            EXTENSION_VALUE_FIELD_NUMBER: _ClassVar[int]
            name: str
            type: int
            int_value: int
            long_value: int
            float_value: float
            double_value: float
            boolean_value: bool
            string_value: str
            extension_value: Payload.Template.Parameter.ParameterValueExtension
            def __init__(
                self,
                name: _Optional[str] = ...,
                type: _Optional[int] = ...,
                int_value: _Optional[int] = ...,
                long_value: _Optional[int] = ...,
                float_value: _Optional[float] = ...,
                double_value: _Optional[float] = ...,
                boolean_value: bool = ...,
                string_value: _Optional[str] = ...,
                extension_value: _Optional[
                    _Union[Payload.Template.Parameter.ParameterValueExtension, _Mapping]
                ] = ...,
            ) -> None: ...
        VERSION_FIELD_NUMBER: _ClassVar[int]
        METRICS_FIELD_NUMBER: _ClassVar[int]
        PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        TEMPLATE_REF_FIELD_NUMBER: _ClassVar[int]
        IS_DEFINITION_FIELD_NUMBER: _ClassVar[int]
        version: str
        metrics: _containers.RepeatedCompositeFieldContainer[Payload.Metric]
        parameters: _containers.RepeatedCompositeFieldContainer[
            Payload.Template.Parameter
        ]
        template_ref: str
        is_definition: bool
        def __init__(
            self,
            version: _Optional[str] = ...,
            metrics: _Optional[_Iterable[_Union[Payload.Metric, _Mapping]]] = ...,
            parameters: _Optional[
                _Iterable[_Union[Payload.Template.Parameter, _Mapping]]
            ] = ...,
            template_ref: _Optional[str] = ...,
            is_definition: bool = ...,
        ) -> None: ...

    class DataSet(_message.Message):
        __slots__ = ("num_of_columns", "columns", "types", "rows")
        Extensions: _python_message._ExtensionDict

        class DataSetValue(_message.Message):
            __slots__ = (
                "int_value",
                "long_value",
                "float_value",
                "double_value",
                "boolean_value",
                "string_value",
                "extension_value",
            )

            class DataSetValueExtension(_message.Message):
                __slots__ = ()
                Extensions: _python_message._ExtensionDict
                def __init__(self) -> None: ...
            INT_VALUE_FIELD_NUMBER: _ClassVar[int]
            LONG_VALUE_FIELD_NUMBER: _ClassVar[int]
            FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
            DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
            BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
            STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
            EXTENSION_VALUE_FIELD_NUMBER: _ClassVar[int]
            int_value: int
            long_value: int
            float_value: float
            double_value: float
            boolean_value: bool
            string_value: str
            extension_value: Payload.DataSet.DataSetValue.DataSetValueExtension
            def __init__(
                self,
                int_value: _Optional[int] = ...,
                long_value: _Optional[int] = ...,
                float_value: _Optional[float] = ...,
                double_value: _Optional[float] = ...,
                boolean_value: bool = ...,
                string_value: _Optional[str] = ...,
                extension_value: _Optional[
                    _Union[Payload.DataSet.DataSetValue.DataSetValueExtension, _Mapping]
                ] = ...,
            ) -> None: ...

        class Row(_message.Message):
            __slots__ = ("elements",)
            Extensions: _python_message._ExtensionDict
            ELEMENTS_FIELD_NUMBER: _ClassVar[int]
            elements: _containers.RepeatedCompositeFieldContainer[
                Payload.DataSet.DataSetValue
            ]
            def __init__(
                self,
                elements: _Optional[
                    _Iterable[_Union[Payload.DataSet.DataSetValue, _Mapping]]
                ] = ...,
            ) -> None: ...
        NUM_OF_COLUMNS_FIELD_NUMBER: _ClassVar[int]
        COLUMNS_FIELD_NUMBER: _ClassVar[int]
        TYPES_FIELD_NUMBER: _ClassVar[int]
        ROWS_FIELD_NUMBER: _ClassVar[int]
        num_of_columns: int
        columns: _containers.RepeatedScalarFieldContainer[str]
        types: _containers.RepeatedScalarFieldContainer[int]
        rows: _containers.RepeatedCompositeFieldContainer[Payload.DataSet.Row]
        def __init__(
            self,
            num_of_columns: _Optional[int] = ...,
            columns: _Optional[_Iterable[str]] = ...,
            types: _Optional[_Iterable[int]] = ...,
            rows: _Optional[_Iterable[_Union[Payload.DataSet.Row, _Mapping]]] = ...,
        ) -> None: ...

    class PropertyValue(_message.Message):
        __slots__ = (
            "type",
            "is_null",
            "int_value",
            "long_value",
            "float_value",
            "double_value",
            "boolean_value",
            "string_value",
            "propertyset_value",
            "propertysets_value",
            "extension_value",
        )

        class PropertyValueExtension(_message.Message):
            __slots__ = ()
            Extensions: _python_message._ExtensionDict
            def __init__(self) -> None: ...
        TYPE_FIELD_NUMBER: _ClassVar[int]
        IS_NULL_FIELD_NUMBER: _ClassVar[int]
        INT_VALUE_FIELD_NUMBER: _ClassVar[int]
        LONG_VALUE_FIELD_NUMBER: _ClassVar[int]
        FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
        DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
        BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
        STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
        PROPERTYSET_VALUE_FIELD_NUMBER: _ClassVar[int]
        PROPERTYSETS_VALUE_FIELD_NUMBER: _ClassVar[int]
        EXTENSION_VALUE_FIELD_NUMBER: _ClassVar[int]
        type: int
        is_null: bool
        int_value: int
        long_value: int
        float_value: float
        double_value: float
        boolean_value: bool
        string_value: str
        propertyset_value: Payload.PropertySet
        propertysets_value: Payload.PropertySetList
        extension_value: Payload.PropertyValue.PropertyValueExtension
        def __init__(
            self,
            type: _Optional[int] = ...,
            is_null: bool = ...,
            int_value: _Optional[int] = ...,
            long_value: _Optional[int] = ...,
            float_value: _Optional[float] = ...,
            double_value: _Optional[float] = ...,
            boolean_value: bool = ...,
            string_value: _Optional[str] = ...,
            propertyset_value: _Optional[_Union[Payload.PropertySet, _Mapping]] = ...,
            propertysets_value: _Optional[
                _Union[Payload.PropertySetList, _Mapping]
            ] = ...,
            extension_value: _Optional[
                _Union[Payload.PropertyValue.PropertyValueExtension, _Mapping]
            ] = ...,
        ) -> None: ...

    class PropertySet(_message.Message):
        __slots__ = ("keys", "values")
        Extensions: _python_message._ExtensionDict
        KEYS_FIELD_NUMBER: _ClassVar[int]
        VALUES_FIELD_NUMBER: _ClassVar[int]
        keys: _containers.RepeatedScalarFieldContainer[str]
        values: _containers.RepeatedCompositeFieldContainer[Payload.PropertyValue]
        def __init__(
            self,
            keys: _Optional[_Iterable[str]] = ...,
            values: _Optional[_Iterable[_Union[Payload.PropertyValue, _Mapping]]] = ...,
        ) -> None: ...

    class PropertySetList(_message.Message):
        __slots__ = ("propertyset",)
        Extensions: _python_message._ExtensionDict
        PROPERTYSET_FIELD_NUMBER: _ClassVar[int]
        propertyset: _containers.RepeatedCompositeFieldContainer[Payload.PropertySet]
        def __init__(
            self,
            propertyset: _Optional[
                _Iterable[_Union[Payload.PropertySet, _Mapping]]
            ] = ...,
        ) -> None: ...

    class MetaData(_message.Message):
        __slots__ = (
            "is_multi_part",
            "content_type",
            "size",
            "seq",
            "file_name",
            "file_type",
            "md5",
            "description",
        )
        Extensions: _python_message._ExtensionDict
        IS_MULTI_PART_FIELD_NUMBER: _ClassVar[int]
        CONTENT_TYPE_FIELD_NUMBER: _ClassVar[int]
        SIZE_FIELD_NUMBER: _ClassVar[int]
        SEQ_FIELD_NUMBER: _ClassVar[int]
        FILE_NAME_FIELD_NUMBER: _ClassVar[int]
        FILE_TYPE_FIELD_NUMBER: _ClassVar[int]
        MD5_FIELD_NUMBER: _ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
        is_multi_part: bool
        content_type: str
        size: int
        seq: int
        file_name: str
        file_type: str
        md5: str
        description: str
        def __init__(
            self,
            is_multi_part: bool = ...,
            content_type: _Optional[str] = ...,
            size: _Optional[int] = ...,
            seq: _Optional[int] = ...,
            file_name: _Optional[str] = ...,
            file_type: _Optional[str] = ...,
            md5: _Optional[str] = ...,
            description: _Optional[str] = ...,
        ) -> None: ...

    class Metric(_message.Message):
        __slots__ = (
            "name",
            "alias",
            "timestamp",
            "datatype",
            "is_historical",
            "is_transient",
            "is_null",
            "metadata",
            "properties",
            "int_value",
            "long_value",
            "float_value",
            "double_value",
            "boolean_value",
            "string_value",
            "bytes_value",
            "dataset_value",
            "template_value",
            "extension_value",
        )

        class MetricValueExtension(_message.Message):
            __slots__ = ()
            Extensions: _python_message._ExtensionDict
            def __init__(self) -> None: ...
        NAME_FIELD_NUMBER: _ClassVar[int]
        ALIAS_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        DATATYPE_FIELD_NUMBER: _ClassVar[int]
        IS_HISTORICAL_FIELD_NUMBER: _ClassVar[int]
        IS_TRANSIENT_FIELD_NUMBER: _ClassVar[int]
        IS_NULL_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        PROPERTIES_FIELD_NUMBER: _ClassVar[int]
        INT_VALUE_FIELD_NUMBER: _ClassVar[int]
        LONG_VALUE_FIELD_NUMBER: _ClassVar[int]
        FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
        DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
        BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
        STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
        BYTES_VALUE_FIELD_NUMBER: _ClassVar[int]
        DATASET_VALUE_FIELD_NUMBER: _ClassVar[int]
        TEMPLATE_VALUE_FIELD_NUMBER: _ClassVar[int]
        EXTENSION_VALUE_FIELD_NUMBER: _ClassVar[int]
        name: str
        alias: int
        timestamp: int
        datatype: int
        is_historical: bool
        is_transient: bool
        is_null: bool
        metadata: Payload.MetaData
        properties: Payload.PropertySet
        int_value: int
        long_value: int
        float_value: float
        double_value: float
        boolean_value: bool
        string_value: str
        bytes_value: bytes
        dataset_value: Payload.DataSet
        template_value: Payload.Template
        extension_value: Payload.Metric.MetricValueExtension
        def __init__(
            self,
            name: _Optional[str] = ...,
            alias: _Optional[int] = ...,
            timestamp: _Optional[int] = ...,
            datatype: _Optional[int] = ...,
            is_historical: bool = ...,
            is_transient: bool = ...,
            is_null: bool = ...,
            metadata: _Optional[_Union[Payload.MetaData, _Mapping]] = ...,
            properties: _Optional[_Union[Payload.PropertySet, _Mapping]] = ...,
            int_value: _Optional[int] = ...,
            long_value: _Optional[int] = ...,
            float_value: _Optional[float] = ...,
            double_value: _Optional[float] = ...,
            boolean_value: bool = ...,
            string_value: _Optional[str] = ...,
            bytes_value: _Optional[bytes] = ...,
            dataset_value: _Optional[_Union[Payload.DataSet, _Mapping]] = ...,
            template_value: _Optional[_Union[Payload.Template, _Mapping]] = ...,
            extension_value: _Optional[
                _Union[Payload.Metric.MetricValueExtension, _Mapping]
            ] = ...,
        ) -> None: ...
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    SEQ_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    metrics: _containers.RepeatedCompositeFieldContainer[Payload.Metric]
    seq: int
    uuid: str
    body: bytes
    def __init__(
        self,
        timestamp: _Optional[int] = ...,
        metrics: _Optional[_Iterable[_Union[Payload.Metric, _Mapping]]] = ...,
        seq: _Optional[int] = ...,
        uuid: _Optional[str] = ...,
        body: _Optional[bytes] = ...,
    ) -> None: ...



from aiospb.messages import EncodingError, Payload, PayloadEncoder
from aiospb.shared import DataType, Metric, ValueType

from . import sparkplug_pb2 as pb2


class _ValueConversor:
    _VALUES = {
        "String": "string_value",
        "UInt8": "int_value",
        "UInt16": "int_value",
        "UInt32": "int_value",
        "UInt32": "int_value",
        "UInt64": "long_value",
        "Float": "float_value",
        "Double": "double_value",
        "Boolean": "boolean_value",
        "Bytes": "bytes_value",
    }

    def save_value(
        self,
        message: pb2.Payload.Metric | pb2.Payload.PropertyValue,
        type_str: str,
        value: ValueType,
    ):
        try:
            if value is None:
                setattr(message, "is_null", True)
            elif type_str == "Int64":
                setattr(message, "long_value", int(value) & 0xFFFFFFFFFFFFFFFF)
            elif type_str == "UInt64":
                setattr(message, "long_value", int(value) & 0xFFFFFFFFFFFFFFFF)
            elif type_str.startswith("Int"):
                setattr(message, "int_value", int(value) & 0xFFFFFFFF)
            elif type_str.startswith("UInt"):
                setattr(message, "int_value", int(value) & 0xFFFFFFFF)
            else:
                setattr(message, self._VALUES[type_str], value)
        except KeyError:
            raise EncodingError(f"Not possible to save value {value} at {message}")
        except ValueError:
            raise EncodingError(f"Not possible to save value {value} as integer")

    def load_value(self, message, type_str: str) -> ValueType:
        if message.HasField("is_null"):
            return
        if type_str == "Int64":
            value = getattr(message, "long_value")
            return value - 0x10000000000000000 if value > 0x7FFFFFFFFFFFFFFF else value
        if type_str == "UInt64":
            return getattr(message, "long_value")
        elif type_str.startswith("Int"):
            value = getattr(message, "int_value")
            return value - 0x100000000 if value > 0x7FFFFFFF else value
        else:
            return getattr(message, self._VALUES[type_str])


class ProtobufPayloadEncoder(PayloadEncoder):
    def __init__(self):
        self._conversor = _ValueConversor()

    def encode(self, payload: Payload) -> bytes:
        """Convert a message to a payload"""
        pb_payload = pb2.Payload()

        try:
            pb_payload.timestamp = payload.timestamp
            if payload.seq is not None:
                pb_payload.seq = payload.seq
            for metric in payload.metrics:
                pb_metric = pb_payload.metrics.add()
                pb_metric.datatype = metric.data_type.value
                pb_metric.timestamp = metric.timestamp
                self._conversor.save_value(
                    pb_metric, metric.data_type.name, metric.value
                )
                if metric.name:
                    pb_metric.name = metric.name
                if metric.alias:
                    pb_metric.alias = metric.alias
                if metric.is_historical:
                    pb_metric.is_historical = True
                if metric.is_transient:
                    pb_metric.is_transient = True
                if metric.properties:
                    values = []
                    for prop in metric.properties.values():
                        if prop.value is None:
                            value = pb2.Payload.PropertyValue(
                                type=prop.data_type.value, is_null=True
                            )
                        else:
                            value = pb2.Payload.PropertyValue(type=prop.data_type.value)
                            self._conversor.save_value(
                                value, prop.data_type.name, prop.value
                            )
                        values.append(value)

                    pb_metric.properties.CopyFrom(
                        pb2.Payload.PropertySet(
                            keys=tuple(metric.properties.keys()), values=values
                        )
                    )

            return pb_payload.SerializeToString()
        except Exception as e:
            raise EncodingError("Error when encoding") from e

    def _decode_metric(self, spb_metric: pb2.Payload.Metric) -> Metric:
        dump = {
            "timestamp": spb_metric.timestamp,
            "dataType": DataType(spb_metric.datatype).name,
        }
        if spb_metric.HasField("is_null"):
            value = None
        else:
            value = self._conversor.load_value(
                spb_metric, DataType(spb_metric.datatype).name
            )
        dump["value"] = value

        if spb_metric.HasField("name"):
            dump["name"] = spb_metric.name
        if spb_metric.HasField("alias"):
            dump["alias"] = spb_metric.alias
        if spb_metric.HasField("is_historical"):
            dump["is_historical"] = spb_metric.is_historical
        if spb_metric.HasField("is_transient"):
            dump["is_transient"] = spb_metric.is_transient
        if spb_metric.HasField("properties"):
            properties = spb_metric.properties
            dump["properties"] = {
                key: {
                    "dataType": DataType(prop.type),
                    "value": self._conversor.load_value(prop, DataType(prop.type).name),
                }
                for key, prop in zip(properties.keys, properties.values)
            }
        return Metric.from_dict(dump)

    def decode(self, b: bytes) -> Payload:
        """Convert payload to a message object"""
        try:
            pb_payload = pb2.Payload.FromString(b)
        except Exception as e:
            raise EncodingError("Error when decoding") from e

        data = {"timestamp": pb_payload.timestamp, "metrics": []}
        if pb_payload.HasField("seq"):
            data["seq"] = pb_payload.seq
        for spb_metric in pb_payload.metrics:
            data["metrics"].append(self._decode_metric(spb_metric))

        return Payload(**data)



import abc
import asyncio

from aiospb.hosts.values import (
    CommandRequest,
    CommandResponse,
    MinMetricInfo,
    WritingResolution,
)
from aiospb.messages import Event, Message
from aiospb.shared import Clock, DeviceKey, Metric, Quality, UtcClock


class MinMetricInfoCache(abc.ABC):
    """Interface to persist aliases and data_type of metrics to be constructed from WritingRequest"""

    @abc.abstractmethod
    async def get(self, key: DeviceKey, names: list[str]) -> list[MinMetricInfo]:
        ...

    @abc.abstractmethod
    async def set(self, key: DeviceKey, info: dict[str, MinMetricInfo]):
        ...

    @abc.abstractmethod
    async def remove(self, key: DeviceKey):
        ...


class MetricsCreator:
    """Create a command with compressed metrics from a CommandRequest"""

    def __init__(self, cache: MinMetricInfoCache, clock: Clock | None = None):
        self._cache = cache
        self._clock = clock or UtcClock()

    async def update_metric_pars(self, msg: Message):
        """Update the parameters cache to construct commands"""
        if msg.event in (Event.DEVICE_IS_BORN, Event.NODE_IS_BORN):
            await self._cache.set(
                msg.key,
                {m.name: MinMetricInfo(m.data_type, m.alias) for m in msg.metrics},
            )

        if msg.event in (Event.DEVICE_HAS_DIED, Event.NODE_HAS_DIED):
            await self._cache.remove(msg.key)

    async def construct_writing_metrics(self, request: CommandRequest) -> list[Metric]:
        metric_pars = await self._cache.get(
            request.key, [req.metric_name for req in request.requests]
        )
        ts = self._clock.timestamp()
        return [
            req.construct_compresed_metric(ts, pars)
            for req, pars in zip(request.requests, metric_pars)
        ]


class _CommandPipe:
    def __init__(self, metrics: list[Metric], timeout: float):
        self._metrics = metrics
        self._timeout = timeout
        self._event = asyncio.Event()
        self._resolutions = [WritingResolution.WritingTimeout] * len(metrics)

    def device_has_died(self):
        self._resolutions = [WritingResolution.EdgeIsOffline] * len(self._metrics)
        self._event.set()

    def is_executed(self, metrics: list[Metric]) -> bool:
        if len(metrics) != len(self._metrics):
            return False

        resolutions = []
        for dm, cm in zip(metrics, self._metrics):
            if dm.name and dm.name != cm.name:
                return False

            if dm.alias and dm.alias != cm.alias:
                return False

            if dm.value != cm.value:
                return False

            res = WritingResolution.GoodWriting
            match dm.quality:
                case Quality.BAD:
                    res = WritingResolution.BadWriting
                case Quality.STALE:
                    res = WritingResolution.StaleWriting

            resolutions.append(res)
        self._resolutions = resolutions
        self._event.set()
        return True

    async def finish(self) -> list[WritingResolution]:
        try:
            await asyncio.wait_for(self._event.wait(), self._timeout)
        except TimeoutError:
            pass

        return self._resolutions


class CommandFollower:
    def __init__(self, clock: Clock | None = None):
        self._clock = clock or UtcClock()

        self._pendings = {}

    async def update_command_pipes(self, msg: Message):
        if msg.event in (Event.DEVICE_IS_BORN, Event.NODE_IS_BORN):
            self._pendings[msg.key] = []

        if msg.event in (Event.DEVICE_DATA_HAS_CHANGED, Event.NODE_DATA_HAS_CHANGED):
            for p in self._pendings.get(msg.key, []):
                if p.is_executed(msg.metrics):
                    break

        if msg.event in (Event.DEVICE_HAS_DIED, Event.NODE_HAS_DIED):
            for pipe in self._pendings.pop(msg.key, []):
                pipe.device_has_died()

    async def follow_command(
        self, key: DeviceKey, metrics: list[Metric], timeout: float = 60.0
    ) -> CommandResponse:
        pipe = _CommandPipe(metrics, timeout)
        if key not in self._pendings:
            return CommandResponse(
                self._clock.timestamp(),
                [WritingResolution.EdgeIsOffline] * len(metrics),
            )

        self._pendings[key].append(pipe)

        resolutions = await pipe.finish()
        if key in self._pendings:
            self._pendings[key].remove(pipe)

        return CommandResponse(self._clock.timestamp(), resolutions)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sparkplug.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 27, 3, "", "sparkplug.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0fsparkplug.proto\x12!com.cirruslink.sparkplug.protobuf"\xf6\x16\n\x07Payload\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x42\n\x07metrics\x18\x02 \x03(\x0b\x32\x31.com.cirruslink.sparkplug.protobuf.Payload.Metric\x12\x0b\n\x03seq\x18\x03 \x01(\x04\x12\x0c\n\x04uuid\x18\x04 \x01(\t\x12\x0c\n\x04\x62ody\x18\x05 \x01(\x0c\x1a\xbe\x04\n\x08Template\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x42\n\x07metrics\x18\x02 \x03(\x0b\x32\x31.com.cirruslink.sparkplug.protobuf.Payload.Metric\x12Q\n\nparameters\x18\x03 \x03(\x0b\x32=.com.cirruslink.sparkplug.protobuf.Payload.Template.Parameter\x12\x14\n\x0ctemplate_ref\x18\x04 \x01(\t\x12\x15\n\ris_definition\x18\x05 \x01(\x08\x1a\xd2\x02\n\tParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\r\x12\x13\n\tint_value\x18\x03 \x01(\rH\x00\x12\x14\n\nlong_value\x18\x04 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x05 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x06 \x01(\x01H\x00\x12\x17\n\rboolean_value\x18\x07 \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x08 \x01(\tH\x00\x12p\n\x0f\x65xtension_value\x18\t \x01(\x0b\x32U.com.cirruslink.sparkplug.protobuf.Payload.Template.Parameter.ParameterValueExtensionH\x00\x1a#\n\x17ParameterValueExtension*\x08\x08\x01\x10\x80\x80\x80\x80\x02\x42\x07\n\x05value*\x08\x08\x06\x10\x80\x80\x80\x80\x02\x1a\xaf\x04\n\x07\x44\x61taSet\x12\x16\n\x0enum_of_columns\x18\x01 \x01(\x04\x12\x0f\n\x07\x63olumns\x18\x02 \x03(\t\x12\r\n\x05types\x18\x03 \x03(\r\x12\x44\n\x04rows\x18\x04 \x03(\x0b\x32\x36.com.cirruslink.sparkplug.protobuf.Payload.DataSet.Row\x1a\xb7\x02\n\x0c\x44\x61taSetValue\x12\x13\n\tint_value\x18\x01 \x01(\rH\x00\x12\x14\n\nlong_value\x18\x02 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12\x17\n\rboolean_value\x18\x05 \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x06 \x01(\tH\x00\x12p\n\x0f\x65xtension_value\x18\x07 \x01(\x0b\x32U.com.cirruslink.sparkplug.protobuf.Payload.DataSet.DataSetValue.DataSetValueExtensionH\x00\x1a!\n\x15\x44\x61taSetValueExtension*\x08\x08\x01\x10\x80\x80\x80\x80\x02\x42\x07\n\x05value\x1a\x62\n\x03Row\x12Q\n\x08\x65lements\x18\x01 \x03(\x0b\x32?.com.cirruslink.sparkplug.protobuf.Payload.DataSet.DataSetValue*\x08\x08\x02\x10\x80\x80\x80\x80\x02*\x08\x08\x05\x10\x80\x80\x80\x80\x02\x1a\x81\x04\n\rPropertyValue\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\x0f\n\x07is_null\x18\x02 \x01(\x08\x12\x13\n\tint_value\x18\x03 \x01(\rH\x00\x12\x14\n\nlong_value\x18\x04 \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x05 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x06 \x01(\x01H\x00\x12\x17\n\rboolean_value\x18\x07 \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x08 \x01(\tH\x00\x12S\n\x11propertyset_value\x18\t \x01(\x0b\x32\x36.com.cirruslink.sparkplug.protobuf.Payload.PropertySetH\x00\x12X\n\x12propertysets_value\x18\n \x01(\x0b\x32:.com.cirruslink.sparkplug.protobuf.Payload.PropertySetListH\x00\x12j\n\x0f\x65xtension_value\x18\x0b \x01(\x0b\x32O.com.cirruslink.sparkplug.protobuf.Payload.PropertyValue.PropertyValueExtensionH\x00\x1a"\n\x16PropertyValueExtension*\x08\x08\x01\x10\x80\x80\x80\x80\x02\x42\x07\n\x05value\x1ao\n\x0bPropertySet\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12H\n\x06values\x18\x02 \x03(\x0b\x32\x38.com.cirruslink.sparkplug.protobuf.Payload.PropertyValue*\x08\x08\x03\x10\x80\x80\x80\x80\x02\x1ah\n\x0fPropertySetList\x12K\n\x0bpropertyset\x18\x01 \x03(\x0b\x32\x36.com.cirruslink.sparkplug.protobuf.Payload.PropertySet*\x08\x08\x02\x10\x80\x80\x80\x80\x02\x1a\xa4\x01\n\x08MetaData\x12\x15\n\ris_multi_part\x18\x01 \x01(\x08\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\t\x12\x0c\n\x04size\x18\x03 \x01(\x04\x12\x0b\n\x03seq\x18\x04 \x01(\x04\x12\x11\n\tfile_name\x18\x05 \x01(\t\x12\x11\n\tfile_type\x18\x06 \x01(\t\x12\x0b\n\x03md5\x18\x07 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x08 \x01(\t*\x08\x08\t\x10\x80\x80\x80\x80\x02\x1a\xe7\x05\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\x04\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x10\n\x08\x64\x61tatype\x18\x04 \x01(\r\x12\x15\n\ris_historical\x18\x05 \x01(\x08\x12\x14\n\x0cis_transient\x18\x06 \x01(\x08\x12\x0f\n\x07is_null\x18\x07 \x01(\x08\x12\x45\n\x08metadata\x18\x08 \x01(\x0b\x32\x33.com.cirruslink.sparkplug.protobuf.Payload.MetaData\x12J\n\nproperties\x18\t \x01(\x0b\x32\x36.com.cirruslink.sparkplug.protobuf.Payload.PropertySet\x12\x13\n\tint_value\x18\n \x01(\rH\x00\x12\x14\n\nlong_value\x18\x0b \x01(\x04H\x00\x12\x15\n\x0b\x66loat_value\x18\x0c \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\r \x01(\x01H\x00\x12\x17\n\rboolean_value\x18\x0e \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x0f \x01(\tH\x00\x12\x15\n\x0b\x62ytes_value\x18\x10 \x01(\x0cH\x00\x12K\n\rdataset_value\x18\x11 \x01(\x0b\x32\x32.com.cirruslink.sparkplug.protobuf.Payload.DataSetH\x00\x12M\n\x0etemplate_value\x18\x12 \x01(\x0b\x32\x33.com.cirruslink.sparkplug.protobuf.Payload.TemplateH\x00\x12\x61\n\x0f\x65xtension_value\x18\x13 \x01(\x0b\x32\x46.com.cirruslink.sparkplug.protobuf.Payload.Metric.MetricValueExtensionH\x00\x1a \n\x14MetricValueExtension*\x08\x08\x01\x10\x80\x80\x80\x80\x02\x42\x07\n\x05value*\x08\x08\x06\x10\x80\x80\x80\x80\x02\x42\x34\n!com.cirruslink.sparkplug.protobufB\x0fSparkplugBProto'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "sparkplug_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n!com.cirruslink.sparkplug.protobufB\017SparkplugBProto"
    _globals["_PAYLOAD"]._serialized_start = 55
    _globals["_PAYLOAD"]._serialized_end = 2989
    _globals["_PAYLOAD_TEMPLATE"]._serialized_start = 195
    _globals["_PAYLOAD_TEMPLATE"]._serialized_end = 769
    _globals["_PAYLOAD_TEMPLATE_PARAMETER"]._serialized_start = 421
    _globals["_PAYLOAD_TEMPLATE_PARAMETER"]._serialized_end = 759
    _globals[
        "_PAYLOAD_TEMPLATE_PARAMETER_PARAMETERVALUEEXTENSION"
    ]._serialized_start = 715
    _globals[
        "_PAYLOAD_TEMPLATE_PARAMETER_PARAMETERVALUEEXTENSION"
    ]._serialized_end = 750
    _globals["_PAYLOAD_DATASET"]._serialized_start = 772
    _globals["_PAYLOAD_DATASET"]._serialized_end = 1331
    _globals["_PAYLOAD_DATASET_DATASETVALUE"]._serialized_start = 910
    _globals["_PAYLOAD_DATASET_DATASETVALUE"]._serialized_end = 1221
    _globals[
        "_PAYLOAD_DATASET_DATASETVALUE_DATASETVALUEEXTENSION"
    ]._serialized_start = 1179
    _globals[
        "_PAYLOAD_DATASET_DATASETVALUE_DATASETVALUEEXTENSION"
    ]._serialized_end = 1212
    _globals["_PAYLOAD_DATASET_ROW"]._serialized_start = 1223
    _globals["_PAYLOAD_DATASET_ROW"]._serialized_end = 1321
    _globals["_PAYLOAD_PROPERTYVALUE"]._serialized_start = 1334
    _globals["_PAYLOAD_PROPERTYVALUE"]._serialized_end = 1847
    _globals["_PAYLOAD_PROPERTYVALUE_PROPERTYVALUEEXTENSION"]._serialized_start = 1804
    _globals["_PAYLOAD_PROPERTYVALUE_PROPERTYVALUEEXTENSION"]._serialized_end = 1838
    _globals["_PAYLOAD_PROPERTYSET"]._serialized_start = 1849
    _globals["_PAYLOAD_PROPERTYSET"]._serialized_end = 1960
    _globals["_PAYLOAD_PROPERTYSETLIST"]._serialized_start = 1962
    _globals["_PAYLOAD_PROPERTYSETLIST"]._serialized_end = 2066
    _globals["_PAYLOAD_METADATA"]._serialized_start = 2069
    _globals["_PAYLOAD_METADATA"]._serialized_end = 2233
    _globals["_PAYLOAD_METRIC"]._serialized_start = 2236
    _globals["_PAYLOAD_METRIC"]._serialized_end = 2979
    _globals["_PAYLOAD_METRIC_METRICVALUEEXTENSION"]._serialized_start = 2938
    _globals["_PAYLOAD_METRIC_METRICVALUEEXTENSION"]._serialized_end = 2970
# @@protoc_insertion_point(module_scope)
